import asyncio
import socket
import logging

log = logging.getLogger(__name__)

UDP_DST_PORT = 33333
RCV_BUFSIZ = 1024

GRACE_SECONDS = 2

DISCOVERY_MSG = b"\x00\x00\x00\x1a"


class DiscoveredDevice:
    @staticmethod
    def parse_UDP_discovery_info(discovery_info):
        data = {}
        if discovery_info[3] != int.from_bytes(b"\x1b"):
            raise AttributeError("Not device info")
        data["ssid"] = discovery_info[8:20].decode("utf-8").replace("\x00", "").strip()
        data["mac"] = discovery_info[20:26].hex(":")
        data["tcp"] = int.from_bytes(discovery_info[26:27])
        data["model"] = (
            discovery_info[28:42].decode("utf-8").replace("\x00", "").strip()
        )
        data["serial"] = (
            discovery_info[43:51].decode("utf-8").replace("\x00", "").strip()
        )
        data["modbus_address"] = discovery_info[136]
        data["sw_ver"] = discovery_info[52] / 100
        data["hw_ver"] = chr(discovery_info[54])
        try:
            data["description"] = (
                discovery_info[55:94].decode("utf-8").replace("\x00", "").strip()
            )
            data["location"] = (
                discovery_info[95:134].decode("utf-8").replace("\x00", "").strip()
            )
        except:
            pass
        return data

    def __init__(self, ip, port, basic_info_string):
        self.values = {}
        self.values["ip_address"] = ip
        self.values["port"] = port
        self.values.update(self.parse_UDP_discovery_info(basic_info_string))

    def __getattr__(self, name):
        return self.values[name]


class Discovery:
    def __init__(self):
        self.loop = asyncio.get_event_loop()
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.sock.bind(("", 0))
        self.sock.settimeout(GRACE_SECONDS)

        self.cancel_event = asyncio.Event()
        self.dev = {}

    async def broadcast_udp_packet(self, ip_address):
        try:
            log.debug(f"Sending Broadcast to {ip_address}")
            await self.loop.sock_sendto(
                self.sock, DISCOVERY_MSG, (ip_address, UDP_DST_PORT)
            )
        except Exception as e:
            log.error(f"Error broadcasting to {ip_address}: {e}")
            raise e

    async def poll(self, broadcast_addresses, stop_if_found=None):
        try:

            # Broadcast UDP packets to all IPs concurrently
            await asyncio.gather(
                *[
                    self.broadcast_udp_packet(ip_address)
                    for ip_address in broadcast_addresses
                ]
            )

            start_time = self.loop.time()
            while (
                not self.cancel_event.is_set()
                and (self.loop.time() - start_time) < GRACE_SECONDS
            ):
                try:
                    # Use asyncio.wait_for to set a timeout for sock_recvfrom
                    data, addr = await asyncio.wait_for(
                        self.loop.sock_recvfrom(self.sock, RCV_BUFSIZ), timeout=1.0
                    )
                    device = DiscoveredDevice(addr[0], addr[1], data)
                    log.info(
                        f"Found device {device.model} {device.serial} {device.ip_address}"
                    )
                    new_mac = device.mac
                    self.dev[new_mac] = device

                    if (
                        stop_if_found is not None
                        and device.serial.lower() == stop_if_found.lower()
                    ):
                        return list(self.dev.values())

                except asyncio.TimeoutError:
                    # Handle timeout and continue the loop
                    continue

                except ValueError as ve:
                    log.error(f"Error parsing discovery info: {ve}")
                    raise

        except Exception as e:
            log.error(f"Error during discovery: {e}")
            raise

        return list(self.dev.values())

    async def get_devices(self, broadcast_addresses):
        try:
            devices = await self.poll(broadcast_addresses)
            return devices
        except Exception as e:
            log.error(f"Error getting devices: {e}")
            raise

    async def get_serial(self, broadcast_addresses, serial):
        try:
            devices = await self.poll(broadcast_addresses, serial)
            for device in devices:
                if device.serial.lower() == serial.lower():
                    return device
            return None
        except Exception as e:
            log.error(f"Error getting serial: {e}")
            raise



import logging
import time
import asyncio
import struct
import re

from .BaseDevice import Device
from ..adapters import RestAPI, Modbus
from ..helper import (
    IntervalMeasurementStats,
    MeasurementType,
    ModbusMapper,
    Measurements,
    Time_Blocks_Measurements,
    Time_Block,
    Measurement,
    Phase_Measurements,
    Total_Measurements,
    Consumed_Energy,
    Excess_Power,
    Max_15min_Power,
    Active_Power_Measurements,
    Counter,
    Counters,
    get_counter_units,
    get_counter_direction,
    get_counter_type,
)
from ..exceptions import MeasurementTypeNotSupported

log = logging.getLogger(__name__)


class Impact(Device):
    """
    Represents an Impact device.

    Attributes:
        supports_measurements (bool): Indicates whether the device supports measurements.
        supports_counters (bool): Indicates whether the device supports counters.
        fw_version (float): The firmware version of the device.
    """

    DEVICE_PARAMETERS = {
        "IE38": {"phases": 3, "resettable_counters": 16, "non_resettable_counters": 4},
        "IE35": {"phases": 3, "resettable_counters": 16, "non_resettable_counters": 4},
        "IE14": {"phases": 1, "resettable_counters": 8, "non_resettable_counters": 4},
        # Add more models as needed
    }

    supports_measurements = True
    supports_counters = True
    supports_interval_measurements = True

    async def init(self):
        """
        Initializes the Impact device.

        This method retrieves basic information, updates the status, and logs a success message.
        """
        await self.get_basic_info()

        if isinstance(self.adapter, Modbus):
            # Get nominal power
            self.nominal_power = await self.get_used_current_and_voltage()
            # check time blocks support
            self.supports_time_blocks = await self.check_time_blocks_support()  
        elif isinstance(self.adapter, RestAPI):
            self.supports_time_blocks = await self.adapter.check_time_blocks_support()
        await self.update_status()
        log.debug(f"Successfully initialized {self.model} {self.serial}")

    async def get_measurements(
        self, measurement_type: MeasurementType = MeasurementType.ACTUAL_MEASUREMENTS
    ):
        """
        Retrieves measurements from the device.

        Returns:
            dict: A dictionary containing the measurements.
        """
        if (
            measurement_type != MeasurementType.ACTUAL_MEASUREMENTS
            and self.supports_interval_measurements == False
        ):
            raise MeasurementTypeNotSupported(
                f"{measurement_type} is not supported by {self.model}"
            )

        if isinstance(self.adapter, RestAPI):
            log.debug(
                f"Getting measurements from Rest API for {self.model} {self.serial}"
            )
            return await self.adapter.get_measurements()
        elif isinstance(self.adapter, Modbus):
            log.debug(
                f"Getting measurements from Modbus for {self.model} {self.serial}"
            )

            offset = 0
            last_interval_duration = None
            time_since_last_measurement = None
            avg_measurement_counter = None

            # Other measurement type registers are just shifted
            if measurement_type == MeasurementType.AVERAGE_MEASUREMENTS:
                offset = 5400
            elif measurement_type == MeasurementType.MAX_MEASUREMENTS:
                offset = 5500
            elif measurement_type == MeasurementType.MIN_MEASUREMENTS:
                offset = 5600

            interval_stats = None
            data = await self.adapter.read_input_registers(100 + offset, 91)
            mapper = ModbusMapper(data, 100)

            if measurement_type != MeasurementType.ACTUAL_MEASUREMENTS:
                interval_stats = IntervalMeasurementStats()
                interval_data = await self.adapter.read_input_registers(5500, 2)
                interval_stats_mapper = ModbusMapper(interval_data, 100)
                interval_stats.last_interval_duration = (
                    interval_stats_mapper.get_uint16(100) / 10
                )
                interval_stats.time_since_last_measurement = (
                    interval_stats_mapper.get_int16(101) / 10
                )

            phases = []
            for phase in range(self.phases):
                voltage = Measurement(
                    mapper.get_t5(107 + 2 * phase),
                    "V",
                )
                current = Measurement(
                    mapper.get_t5(126 + 2 * phase),
                    "A",
                )
                active_power = Measurement(
                    mapper.get_t6(142 + 2 * phase),
                    "W",
                )
                reactive_power = Measurement(
                    mapper.get_t6(150 + 2 * phase),
                    "var",
                )
                apparent_power = Measurement(
                    mapper.get_t5(158 + 2 * phase),
                    "VA",
                )
                power_factor = Measurement(
                    mapper.get_t7(166 + 2 * phase)["value"],
                    "",
                )
                power_angle = Measurement(
                    mapper.get_int16(173 + phase) / 100,
                    "°",
                )
                thd_voltage = Measurement(
                    mapper.get_uint16(182 + phase) / 100,
                    "%",
                )
                thd_current = Measurement(
                    mapper.get_uint16(188 + phase) / 100,
                    "%",
                )
                phases.append(
                    Phase_Measurements(
                        voltage,
                        current,
                        active_power,
                        reactive_power,
                        apparent_power,
                        power_factor,
                        power_angle,
                        thd_voltage,
                        thd_current,
                    )
                )

            active_power_total = Measurement(
                mapper.get_t6(140),
                "W",
            )
            reactive_power_total = Measurement(
                mapper.get_t6(148),
                "var",
            )
            apparent_power_total = Measurement(
                mapper.get_t5(156),
                "VA",
            )
            power_factor_total = Measurement(
                mapper.get_t7(164)["value"],
                "",
            )
            power_angle_total = Measurement(
                mapper.get_int16(172) / 100,
                "°",
            )
            frequency = Measurement(
                mapper.get_t5(105),
                "Hz",
            )
            temperature = Measurement(
                mapper.get_int16(181) / 100,
                "°C",
            )
            total = Total_Measurements(
                active_power_total,
                reactive_power_total,
                apparent_power_total,
                power_factor_total,
                power_angle_total,
            )

            return Measurements(phases, total, frequency, temperature, interval_stats)

    async def get_counters(self):
        """
        Retrieves counters from the device.

        Returns:
            dict: A dictionary containing the counters.
        """
        if isinstance(self.adapter, RestAPI):
            log.debug(f"Getting counters from Rest API for {self.model} {self.serial}")
            return await self.adapter.get_counters()
        elif isinstance(self.adapter, Modbus):
            # Open the connection
            handle_connection = not self.adapter.connected
            if handle_connection:
                await self.adapter.open_connection()

            log.debug(f"Getting counters from Modbus for {self.model} {self.serial}")
            data = await self.adapter.read_input_registers(2750, 96)
            data_mapper = ModbusMapper(data, 2750)

            direction_settings = await self.adapter.read_holding_registers(151, 1)

            non_resettable_counter_settings = await self.adapter.read_holding_registers(
                421, 16
            )
            non_resettable_settings_mapper = ModbusMapper(
                non_resettable_counter_settings, 421
            )

            resettable_counter_settings = await self.adapter.read_holding_registers(
                437, 64
            )
            resettable_settings_mapper = ModbusMapper(resettable_counter_settings, 437)

            if handle_connection:
                await self.adapter.close_connection()

            non_resettable = []
            resettable = []
            reverse_connection = False
            if direction_settings[0] & 2:
                reverse_connection = True

            for counter in range(self.non_resettable_counters):
                units = get_counter_units(non_resettable_settings_mapper.get_uint16(421 + 4 * counter))

                direction = get_counter_direction(
                    non_resettable_settings_mapper.get_uint16(422 + 4 * counter),
                    reverse_connection,
                )

                counter_type = get_counter_type(direction, units)
                non_resettable.append(
                    Counter(
                        data_mapper.get_float(2752 + 2 * counter),
                        units,
                        direction,
                        counter_type,
                    )
                )

            for counter in range(self.resettable_counters):
                units = get_counter_units(resettable_settings_mapper.get_uint16(437 + 4 * counter))

                direction = get_counter_direction(
                    resettable_settings_mapper.get_uint16(438 + 4 * counter),
                    reverse_connection,
                )

                counter_type = get_counter_type(direction, units)
                
                resettable.append(
                    Counter(
                        data_mapper.get_float(2760 + 2 * counter),
                        units,
                        direction,
                        counter_type,
                    )
                )

            return Counters(non_resettable, resettable)

    async def update_status(self):
        """
        Updates the status of the device.

        This method acquires a lock to ensure that only one update is running at a time.
        It retrieves measurements and counters, updates the corresponding attributes,
        and sets the update timestamp.
        """
        # If update is already running, wait for it to finish and then return
        if self.update_lock.locked():
            log.debug("Update already running for %s %s" % (self.model, self.serial))
            while self.update_lock.locked():
                await asyncio.sleep(0.1)
            return

        # If update is not running, acquire the lock and update
        async with self.update_lock:
            log.debug("Updating status for %s %s" % (self.model, self.serial))

            # if the adapter is Modbus, open the connection
            if isinstance(self.adapter, Modbus):
                await self.adapter.open_connection()

            self.measurements = await self.get_measurements()
            self.counters = await self.get_counters()
            # Time blocks measurements
            self.time_blocks_measurements = await self.get_time_blocks_measurements()

            # if the adapter is Modbus, close the connection
            if isinstance(self.adapter, Modbus):
                await self.adapter.close_connection()

            self.update_timestamp = time.time()

    async def get_used_current_and_voltage(self):
        """
        Get meter configuration of used current and voltage.
        """
        log.debug(f"Getting config registers.")
        cnf_data = await self.adapter.read_holding_registers(148, 2)
        mapper = ModbusMapper(cnf_data, 148)

        cnf_data = await self.adapter.read_input_registers(15, 4)
        data_mapper = ModbusMapper(cnf_data, 15) # used voltage and current in cV & cA

        used_current = (mapper.get_uint16(148)/10000) * (data_mapper.get_uint16(17)/100)
        used_voltage = (mapper.get_uint16(149)/10000) * (data_mapper.get_uint16(15)/100)

        nominal_power = 3 * used_voltage * used_current

        return nominal_power
            
    async def get_number_of_time_blocks(self):
        """
        Get meter configuration status.
        """
        log.debug(f"Getting config register status.")
        cnf_data = await self.adapter.read_input_registers(6761, 1)
        time_blocks = ModbusMapper(cnf_data, 6761)
        return time_blocks
    
    async def check_time_blocks_support(self):
        """
        Check Time Blocks support
        """
        mapper = await self.get_number_of_time_blocks()
        number_of_blocks = mapper.get_uint16(6761)
        if number_of_blocks == 5:
            return True
        else:
            return False

    async def get_time_blocks_measurements(self):
        """
        Retrieves time block measurements from the device.

        Returns:
            dict: A dictionary containing the time blocks measurements.
        """

        if isinstance(self.adapter, RestAPI):
            log.debug(f"Getting counters from Rest API for {self.model} {self.serial}")
            if self.supports_time_blocks:
                return await self.adapter.get_tb_measurements()
            else:
                return None
        
        elif isinstance(self.adapter, Modbus):
            if (not self.supports_time_blocks):
                return None
            # Open the connection
            handle_connection = not self.adapter.connected
            if handle_connection:
                await self.adapter.open_connection()

            log.debug(f"Getting time blocks measurements for {self.model} {self.serial}")
            data = await self.adapter.read_input_registers(6761, 238)
            mapper = ModbusMapper(data, 6761)

            data = await self.adapter.read_holding_registers(990, 5)
            limits__mapper = ModbusMapper(data, 990)

            active_power_measurements = await self.adapter.read_input_registers(5244, 18)
            active_power_measurements_mapper = ModbusMapper(active_power_measurements, 5244)

            capture_timestamp = await self.adapter.read_input_registers(4901, 2)
            capture_timestamp_mapper = ModbusMapper(capture_timestamp, 4901)

            pred_15min_active_limit = await self.adapter.read_input_registers(5261, 2)
            pred_15min_active_limit_mapper = ModbusMapper(pred_15min_active_limit, 5261)

            if handle_connection:
                await self.adapter.close_connection()

            time_blocks = []
            consumed_energy = []
            excess_power = []
            max_15min_power = []

            number_of_blocks = mapper.get_uint16(6761)
            exponent = mapper.get_int16(6762)

            for block in range (number_of_blocks):
                total =  Measurement(mapper.get_uint32(6786 + 42 * block)* (10**exponent),
                    "Wh",
                )
                timestamp_total =  Measurement(capture_timestamp_mapper.get_timestamp(4901),
                "",
                )
                last_month =  Measurement(mapper.get_uint32(6788 + 42 * block)* (10**exponent),
                    "Wh",
                )
                timestamp_last_month =  Measurement(mapper.get_timestamp(6766),
                "",
                )
                two_months_ago =  Measurement(mapper.get_uint32(6790 + 42 * block)* (10**exponent),
                    "Wh",
                )
                timestamp_two_months_ago =  Measurement(mapper.get_timestamp(6768),
                "",
                )
                last_year =  Measurement(mapper.get_uint32(6792 + 42 * block)* (10**exponent),
                    "Wh",
                )
                timestamp_last_year =  Measurement(mapper.get_timestamp(6770),
                "",
                )
                two_years_ago =  Measurement(mapper.get_uint32(6794 + 42 * block)* (10**exponent),
                    "Wh",
                )
                timestamp_two_years_ago =  Measurement(mapper.get_timestamp(6772),
                "",
                )
                this_month =  Measurement(mapper.get_uint32(6796 + 42 * block)* (10**exponent),
                    "Wh",
                )
                previous_month =  Measurement(mapper.get_uint32(6798 + 42 * block)* (10**exponent),
                    "Wh",
                )
                this_year =  Measurement(mapper.get_uint32(6800 + 42 * block)* (10**exponent),
                    "Wh",
                )
                previous_year =  Measurement(mapper.get_uint32(6802 + 42 * block)* (10**exponent),
                    "Wh",
                )
                consumed_energy.append(
                    Consumed_Energy(
                        total,
                        timestamp_total,
                        last_month,
                        timestamp_last_month,
                        two_months_ago,
                        timestamp_two_months_ago,
                        last_year,
                        timestamp_last_year,
                        two_years_ago,
                        timestamp_two_years_ago,
                        this_month,
                        previous_month,
                        this_year,
                        previous_year,
                    )
                )
                excess_power_limit =  Measurement(round((limits__mapper.get_uint16(990 + 1 * block) * (self.nominal_power/1000))/10), 
                    "W",
                )
                excess_power_this_month =  Measurement(mapper.get_t5(6804 + 42 * block),
                    "W",
                )
                excess_power_previous_month =  Measurement(mapper.get_t5(6806 + 42 * block),
                    "W",
                )
                excess_power.append(
                    Excess_Power(
                        excess_power_limit,
                        excess_power_this_month,
                        excess_power_previous_month,
                    )
                )
                max_15min_power_since_reset =  Measurement(mapper.get_t5(6808 + 42 * block),
                    "W",
                )
                timestamp_since_reset =  Measurement(mapper.get_timestamp(6810 + 42 * block),
                "",
                )
                max_15min_power_this_month =  Measurement(mapper.get_t5(6812 + 42 * block),
                    "W",
                )
                timestamp_this_month =  Measurement(mapper.get_timestamp(6814 + 42 * block),
                "",
                )
                max_15min_power_previous_month =  Measurement(mapper.get_t5(6816 + 42 * block),
                    "W",
                )
                timestamp_previous_month =  Measurement(mapper.get_timestamp(6818 + 42 * block),
                "",
                )
                max_15min_power_this_year =  Measurement(mapper.get_t5(6820 + 42 * block),
                    "W",
                )
                timestamp_this_year =  Measurement(mapper.get_timestamp(6822 + 42 * block),
                "",
                )
                max_15min_power_previous_year =  Measurement(mapper.get_t5(6824 + 42 * block),
                    "W",
                )
                timestamp_previous_year =  Measurement(mapper.get_timestamp(6826 + 42 * block),
                "",
                )
                reset_timestamp =  Measurement(mapper.get_timestamp(6764),
                "",
                )
                max_15min_power.append(
                    Max_15min_Power(
                        max_15min_power_since_reset,
                        timestamp_since_reset,
                        max_15min_power_this_month,
                        timestamp_this_month,
                        max_15min_power_previous_month,
                        timestamp_previous_month,
                        max_15min_power_this_year,
                        timestamp_this_year,
                        max_15min_power_previous_year,
                        timestamp_previous_year,
                        reset_timestamp,
                    )
                )

                time_blocks.append(
                    Time_Block(
                        consumed_energy,
                        excess_power,
                        max_15min_power,
                    )
                )

            actual_value =  Measurement(active_power_measurements_mapper.get_t5(5245),
                "W",
            )
            thermal_function =  Measurement(active_power_measurements_mapper.get_t5(5247),
                "W",
            )
            predicted_15min =  Measurement(active_power_measurements_mapper.get_t5(5249),
                "W",
            )
            predicted_15min_active_limit =  Measurement(pred_15min_active_limit_mapper.get_int16(5261)/100,
                "%",
            )
            last_15min =  Measurement(active_power_measurements_mapper.get_t5(5251),
                "W",
            )
            max_15min_since_reset =  Measurement(mapper.get_t5(6778),
                "W",
            )
            active_energy_total =  Measurement(mapper.get_uint32(6774)* (10**exponent),
                "W",
            )
            timestamp =  Measurement(mapper.get_timestamp(6780),
                "",
            )
            active_power_measurements_import = Active_Power_Measurements(
                actual_value, 
                thermal_function,
                predicted_15min,
                predicted_15min_active_limit,
                last_15min,
                max_15min_since_reset,
                active_energy_total,
                timestamp,
            )
            
            actual_value =  Measurement(active_power_measurements_mapper.get_t5(5253),
                "W",
            )
            thermal_function =  Measurement(active_power_measurements_mapper.get_t5(5255),
                "W",
            )
            predicted_15min =  Measurement(active_power_measurements_mapper.get_t5(5257),
                "W",
            )
            predicted_15min_active_limit =  Measurement(0,
                "",
            )
            last_15min =  Measurement(active_power_measurements_mapper.get_t5(5259),
                "W",
            )
            max_15min_since_reset =  Measurement(mapper.get_uint32(6782)* (10**exponent),
                "W",
            )
            active_energy_total =  Measurement(mapper.get_uint32(6776)* (10**exponent),
                "W",
            )
            timestamp =  Measurement(mapper.get_timestamp(6784),
                "",
            )
            active_power_measurements_export = Active_Power_Measurements(
                actual_value, 
                thermal_function,
                predicted_15min,
                predicted_15min_active_limit,
                last_15min,
                max_15min_since_reset,
                active_energy_total,
                timestamp,
            )
            
            active_block_index =  Measurement(mapper.get_uint16(6763),
                "",
            ) 
            time_to_end_interval =  Measurement(active_power_measurements_mapper.get_uint16(5244),
                "s",
            )

            return Time_Blocks_Measurements(time_blocks, active_power_measurements_import, active_power_measurements_export, active_block_index, time_to_end_interval)



import aiohttp
import base64
import logging
from aiohttp import ClientConnectionError, ServerTimeoutError
from .BaseAdapter import Adapter
from ..helper import (
    BasicInfo,
    Measurements,
    Measurement,
    Phase_Measurements,
    Total_Measurements,
    Time_Blocks_Measurements,
    Counters,
    Counter,
    Time_Block,
    Consumed_Energy,
    Excess_Power,
    get_counter_type,
)

from ..exceptions import (
    NotAuthorised,
    ProtocolNotSupported,
    InvalidResponseCode,
    DeviceConnectionError,
    DeviceTimeoutError,
)

log = logging.getLogger(__name__)


class RestAPI(Adapter):
    """Adapter class for making REST API calls."""

    def __init__(self, ip_address, device_index=None, authentication=None):
        """
        Initialize the RestAPI adapter.

        Args:
            ip_address (str): The IP address of the REST API.
            device_index (int, optional): The index of the device. Defaults to None.
            authentication (dict, optional): The authentication credentials. Defaults to None.
        """
        self.ip_address = ip_address
        self.device_index = device_index
        self.authentication = authentication

    async def get_resource(self, resource):
        """
        Get a resource from the REST API.

        Args:
            resource (str): The resource path.

        Returns:
            dict: The parsed JSON response.

        Raises:
            Exception: If an error occurs while making the REST API call.
        """
        headers = {}
        if self.authentication:
            authentication = self.authentication

            if (password := authentication.get("password")) and (
                username := authentication.get("username")
            ):
                headers["cookie"] = "Authorization=Basic " + base64.b64encode(
                    (username + ":" + password).encode("utf-8")
                ).decode("utf-8")

        # Set a timeout for the REST API call
        try:
            timeout = aiohttp.ClientTimeout(total=10)
            async with aiohttp.ClientSession(timeout=timeout) as session:
                async with session.get(
                    f"http://{self.ip_address}/{resource}", headers=headers
                ) as response:
                    return await RestAPI.handle_response(response)
        except ServerTimeoutError as e:
            raise DeviceTimeoutError(
                f"Timeout occurred while connecting to the RestAPI: {e}"
            ) from e
        except ClientConnectionError as e:
            raise DeviceConnectionError(
                f"Error occurred while connecting to the RestAPI: {e}"
            ) from e
        except NotAuthorised as e:
            raise NotAuthorised(f"Not authorised to access the device: {e}") from e
        except ProtocolNotSupported as e:
            raise ProtocolNotSupported(f"{e}") from e
        except InvalidResponseCode:
            raise DeviceConnectionError(
                f"Invalid response code while connecting to the RestAPI"
            )
        except Exception as e:
            raise DeviceConnectionError(
                f"Error occurred while connecting to the RestAPI: {e}"
            ) from e

    @staticmethod
    async def handle_response(response):
        """
        Handle the HTTP response.

        Args:
            response: The HTTP response object.

        Returns:
            dict: The parsed JSON response.

        Raises:
            Exception: If the response status is unexpected.
        """
        if response.status == 200:
            return await RestAPI.parse_resource(response)
        elif response.status == 403:
            raise NotAuthorised("Not authorised")
        elif response.status == 404:
            raise ProtocolNotSupported("Device not supported by RestAPI")
        else:
            raise InvalidResponseCode("Invalid response code")

    @staticmethod
    async def parse_resource(response):
        """
        Parse the HTTP response.

        Args:
            response: The HTTP response object.

        Returns:
            dict: The parsed JSON response.
        """
        return await response.json()

    async def get_devices(self):
        """
        Get the list of devices.

        Returns:
            list: The list of devices.
        """
        child_devices = (await self.get_resource("api/devices")).get("devices", [])
        return child_devices

    def parse_measurement(self, measurement, default_unit=None):
        # New format: dict with 'value' and 'unit'
        if (
            isinstance(measurement, dict)
            and "value" in measurement
            and "unit" in measurement
        ):
            return Measurement(float(measurement["value"]), measurement["unit"])
        elif isinstance(measurement, str):
            parts = measurement.split()
            if len(parts) == 2:
                return Measurement(float(parts[0]), parts[1])
            elif len(parts) == 1 and default_unit:
                return Measurement(float(parts[0]), default_unit)

        # Already a float/int, use default unit if provided
        elif isinstance(measurement, (float, int)) and default_unit:
            return Measurement(float(measurement), default_unit)
        return None

    async def get_measurements(self):
        """
        Get the measurements.

        Returns:
            Measurements: The measurements object.

        Raises:
            ConnectionError: If an error occurs while connecting to the RestAPI.
        """
        json_data = None
        if self.device_index is not None and self.device_index >= 0:
            json_data = await self.get_resource(
                "api/measurement/" + str(self.device_index)
            )
            json_data = json_data.get("measurements", {})
            phases = []
            total = None
            frequency = None
            temperature = None

            for phase in json_data.get("Phases", []):
                voltage = None
                current = None
                active_power = None
                reactive_power = None
                apparent_power = None
                power_factor = None
                power_angle = None
                thd_voltage = None
                thd_current = None

                temp = phase.get("U", phase.get("voltage", None))
                if temp:
                    voltage = self.parse_measurement(temp)

                temp = phase.get("I", phase.get("current", None))
                if temp:
                    current = self.parse_measurement(temp)

                temp = phase.get("P", phase.get("active_power", None))
                if temp:
                    active_power = self.parse_measurement(temp)

                temp = phase.get("Q", phase.get("reactive_power", None))
                if temp:
                    reactive_power = self.parse_measurement(temp)

                temp = phase.get("S", phase.get("apparent_power", None))
                if temp:
                    apparent_power = self.parse_measurement(temp)

                temp = phase.get("PF", phase.get("power_factor", None))
                if temp:
                    power_factor = self.parse_measurement(temp)

                temp = phase.get("PA", phase.get("power_angle", None))
                if temp:
                    power_angle = self.parse_measurement(temp, "°")

                temp = phase.get("THDUp", phase.get("THD_voltage", None))
                if temp:
                    thd_voltage = self.parse_measurement(temp, "%")

                temp = phase.get("THDI", phase.get("THD_current", None))
                if temp:
                    thd_current = self.parse_measurement(temp, "%")

                phases.append(
                    Phase_Measurements(
                        voltage,
                        current,
                        active_power,
                        reactive_power,
                        apparent_power,
                        power_factor,
                        power_angle,
                        thd_voltage,
                        thd_current,
                    )
                )
            if json_data.get("Total"):
                total_measurements = json_data.get("Total", {})

                temp = total_measurements.get(
                    "P", total_measurements.get("active_power", None)
                )
                if temp:
                    active_power = self.parse_measurement(temp)

                temp = total_measurements.get(
                    "Q", total_measurements.get("reactive_power", None)
                )
                if temp:
                    reactive_power = self.parse_measurement(temp)

                temp = total_measurements.get(
                    "S", total_measurements.get("apparent_power", None)
                )
                if temp:
                    apparent_power = self.parse_measurement(temp)

                temp = total_measurements.get(
                    "PF", total_measurements.get("power_factor", None)
                )
                if temp:
                    power_factor = self.parse_measurement(temp)

                temp = total_measurements.get(
                    "PA", total_measurements.get("power_angle", None)
                )
                if temp:
                    power_angle = self.parse_measurement(temp, "°")

                total = Total_Measurements(
                    active_power,
                    reactive_power,
                    apparent_power,
                    power_factor,
                    power_angle,
                )

            temp = json_data.get("Frequency", json_data.get("frequency", None))
            if temp:
                frequency = self.parse_measurement(temp, "Hz")

            temp = json_data.get("Temperature", json_data.get("temperature", None))
            if temp:
                temperature = self.parse_measurement(temp, "°C")

            return Measurements(
                phases=phases,
                total=total,
                frequency=frequency,
                temperature=temperature,
            )

        return None

    async def get_counters(self):
        """
        Get the counters.

        Returns:
            tuple: A tuple containing the non-resettable, resettable counters.

        Raises:
            ConnectionError: If an error occurs while connecting to the RestAPI.
        """
        json_data = None

        json_data = await self.get_resource("api/counter/" + str(self.device_index))

        json_data = json_data.get("counters", {})
        non_resettable = []
        resettable = []
        if json_data.get("non_resettable"):
            for counter in json_data.get("non_resettable", []):
                try:
                    conunter_type = get_counter_type(
                        counter.get("direction"), counter.get("unit")
                    )
                    non_resettable.append(
                        Counter(
                            counter.get("value"),
                            counter.get("unit"),
                            counter.get("direction"),
                            conunter_type,
                        )
                    )
                except Exception as e:
                    log.error(
                        "Failed to parse non-resettable counter: %s: %s",
                        counter.get("name"),
                        e,
                    )

        if json_data.get("resettable"):
            for counter in json_data.get("resettable", []):
                try:
                    conunter_type = get_counter_type(
                        counter.get("direction"), counter.get("unit")
                    )
                    resettable.append(
                        Counter(
                            counter.get("value"),
                            counter.get("unit"),
                            counter.get("direction"),
                            conunter_type,
                        )
                    )
                except Exception as e:
                    log.error(
                        "Failed to parse resettable counter: %s: %s",
                        counter.get("name"),
                        e,
                    )
        return Counters(non_resettable, resettable)

    async def get_basic_info(self):
        """
        Retrieves basic information about the device.

        Returns:
            BasicInfo: An object containing the basic information of the device.
        """
        json_data = None
        if self.device_index is not None and self.device_index >= 0:
            try:
                child_devices = await self.get_devices()
            except Exception as e:
                raise DeviceConnectionError(
                    f"Failed to connect to the device: {e}"
                ) from e
            # filter out the right IR device
            child_devices = [
                device
                for device in child_devices
                if "Right IR" not in device["interface"]
            ]
            json_data = child_devices[self.device_index]
        else:

            json_data = (await self.get_resource("api")).get("device", {})
            # fix syntax to be consistent with other devices
            json_data["model"] = json_data["model_type"].strip()
            json_data["serial"] = json_data["serial_number"].strip()
            del json_data["model_type"]
            del json_data["serial_number"]

        return BasicInfo(
            serial=json_data.get("serial"),
            model=json_data.get("model"),
            description=json_data.get("description"),
            location=json_data.get("location"),
            sw_ver=json_data.get("sw_ver"),
        )
    
    async def check_time_blocks_support(self):
        """
        Check if device support time blocks.

        Returns:
            bool: Boolean value of support status.

        Raises:
            ConnectionError: If an error occurs while connecting to the RestAPI.
        """
        
        self.number_of_time_blocks = 0

        json_data = None

        json_data = await self.get_resource("api/counter/" + str(self.device_index))

        json_data = json_data.get("counters", {})

        block_data = json_data.get("block")

        if block_data != None:
            # get number of time blocks from RestAPI
            block = block_data.get("blocks", [])[0]
            for item in block:
                if "time_block" in item:
                    self.number_of_time_blocks += 1            
            return True
        else:
            return False

    
    async def get_tb_measurements(self):
        """
        Get time blocks.

        Returns:
            tuple: A tuple containing the time block measurements.

        Raises:
            ConnectionError: If an error occurs while connecting to the RestAPI.
        """
        json_data = None

        json_data = await self.get_resource("api/counter/" + str(self.device_index))

        json_data = json_data.get("counters", {})

        time_blocks = []

        consumed_energy = []
        excess_power = []
        max_15min_power = None

        active_power_measurements_import = None
        active_power_measurements_export = None
        time_to_end_interval = None

        total = None
        timestamp_total = None
        last_month= None
        timestamp_last_month = None
        two_months_ago = None 
        timestamp_two_months_ago = None
        last_year = None
        timestamp_last_year = None
        two_years_ago = None
        timestamp_two_years_ago = None

        excess_power_limit = None

        if json_data.get("block"):

            block_data = json_data.get("block", {})


            active_block_index = Measurement(
                                                block_data.get("active_block"), 
                                                "",
                                             )
            
        
            try:
                for block_index in range (self.number_of_time_blocks):           
                    for block in block_data.get("blocks", []): 
                        if block.get("name") == "This Month Consumed Energy":  
                            this_month = Measurement(
                                    block.get(f"time_block_{block_index + 1}"),
                                    block.get("unit"),
                            )
                        elif block.get("name") == "Previous Month Consumed Energy":  
                            previous_month = Measurement(
                                    block.get(f"time_block_{block_index + 1}"),
                                    block.get("unit"),
                            )
                        elif block.get("name") == "This Year Consumed Energy":  
                            this_year = Measurement(
                                    block.get(f"time_block_{block_index + 1}"),
                                    block.get("unit"),
                            )
                        elif block.get("name") == "Previous Year Consumed Energy":  
                            previous_year = Measurement(
                                    block.get(f"time_block_{block_index + 1}"),
                                    block.get("unit"),
                            )
                        elif block.get("name") == "This Month Excess Power":  
                            excess_power_this_month = Measurement(
                                    block.get(f"time_block_{block_index + 1}"),
                                    block.get("unit"),
                            )
                        elif block.get("name") == "Previous Month Excess Power":  
                            excess_power_previous_month = Measurement(
                                    block.get(f"time_block_{block_index + 1}"),
                                    block.get("unit"),
                            )
                        else:
                            this_month = Measurement(0, "")
                            previous_month = Measurement(0, "")
                            this_year = Measurement(0, "")
                            previous_year = Measurement(0, "")
                            excess_power_this_month = Measurement(0, "")
                            excess_power_previous_month = Measurement(0, "")



                    consumed_energy.append(
                        Consumed_Energy(
                            total,
                            timestamp_total,
                            last_month,
                            timestamp_last_month,
                            two_months_ago,
                            timestamp_two_months_ago,
                            last_year,
                            timestamp_last_year,
                            two_years_ago,
                            timestamp_two_years_ago,
                            this_month,
                            previous_month,
                            this_year,
                            previous_year,
                        )
                    )
                    excess_power.append(
                        Excess_Power(
                            excess_power_limit,
                            excess_power_this_month,
                            excess_power_previous_month,
                        )
                        )


                    time_blocks.append(
                        Time_Block(
                            consumed_energy,
                            excess_power,
                            max_15min_power,
                        )
                    )

            except Exception as e:
                log.error(
                    "Failed to parse resettable counter: %s: %s",
                    block.get("name"),
                    e,
                )

        return Time_Blocks_Measurements(time_blocks, active_power_measurements_import, active_power_measurements_export, active_block_index, time_to_end_interval)


import logging
from pymodbus.client import AsyncModbusTcpClient
from pymodbus.client import AsyncModbusSerialClient

from .BaseAdapter import Adapter
from ..helper import BasicInfo, ModbusMapper
from ..exceptions import InvalidResponseCode, DeviceConnectionError

log = logging.getLogger(__name__)


class Modbus(Adapter):
    """Adapter class for making REST API calls."""

    def __init__(
        self,
        protocol: str,
        ip_address=None,
        modbus_address=33,
        port=10001,
        stopbits=2,
        bytesize=8,
        parity="N",
        baudrate=115200,
    ):
        """
        Initialize the RestAPI adapter.

        Args:
            ip_address (str): The IP address of the REST API.
            device_index (int, optional): The index of the device. Defaults to None.
            authentication (dict, optional): The authentication credentials. Defaults to None.
        """
        self.modbus_address = modbus_address
        self.ip_address = ip_address
        self.port = port
        if protocol == "tcp":
            self.protocol = "tcp"
            self.client = AsyncModbusTcpClient(host=ip_address, port=port, timeout=1)
        elif protocol == "rtu":
            self.protocol = "rtu"
            self.client = AsyncModbusSerialClient(
                port=port,
                stopbits=stopbits,
                bytesize=bytesize,
                parity=parity,
                baudrate=baudrate,
                timeout=1,
            )
        else:
            raise ValueError("Invalid protocol")

    @staticmethod
    def convert_registers_to_string(registers):
        """Converts a list of 16-bit registers to a string, separating each 8 bits of the register for each character."""
        string = ""
        for register in registers:
            high_byte = register >> 8
            low_byte = register & 0xFF
            string += chr(high_byte) + chr(low_byte)
        return string.split("\0")[0].strip()

    async def open_connection(self):
        """Connects to the device."""
        log.debug(f"Connecting to the device {self.ip_address}")
        if not self.client:
            raise DeviceConnectionError("The connection is not configured")

        await self.client.connect()
        if not self.connected:
            if self.protocol == "tcp":
                raise DeviceConnectionError(
                    f"Failed to connect to the device {self.ip_address} on port {self.port}"
                )
            elif self.protocol == "rtu":
                raise DeviceConnectionError(
                    f"Failed to connect to the device on port {self.port}"
                )

    async def close_connection(self):
        """Closes the connection to the device."""
        log.debug(f"Closing the connection to the device {self.ip_address}")
        if not self.client:
            raise DeviceConnectionError("The connection is not configured")

        self.client.close()
        if self.connected:
            if self.protocol == "tcp":
                raise DeviceConnectionError(
                    f"Failed to close the connection to the device {self.ip_address} on port {self.port}"
                )
            elif self.protocol == "rtu":
                raise DeviceConnectionError(
                    f"Failed to close the connection to the device on port {self.port}"
                )

    @property
    def connected(self) -> bool:
        """Returns the connection status."""
        return self.client.connected

    async def get_basic_info(self):
        """
        Retrieves basic information about the device.

        Returns:
            BasicInfo: An object containing the basic information of the device.

        Raises:
            DeviceConnectionError: If reading failed
        """
        basic_info = {}
        await self.open_connection()

        try:
            try:
                data = await self.read_input_registers(1, 14)
                mapper = ModbusMapper(data, 1)
            except Exception as e:
                raise DeviceConnectionError(
                    f"Failed to read basic information: {e}"
                ) from e

            basic_info["model"] = mapper.get_string_range(1, 8).strip()
            basic_info["serial"] = mapper.get_string_range(9, 4).strip()
            basic_info["sw_ver"] = mapper.get_uint16(13) / 100

            try:
                data = await self.read_holding_registers(101, 40)
                mapper = ModbusMapper(data, 101)
            except Exception as e:
                raise DeviceConnectionError(
                    f"Failed to read description & location: {e}"
                ) from e

            basic_info["description"] = mapper.get_string_range(101, 20)
            basic_info["location"] = mapper.get_string_range(121, 20)

            return BasicInfo(**basic_info)

        finally:
            # Always ensure connection is closed
            await self.close_connection()

    async def read_holding_registers(self, start, count, max_registers_per_read=120):
        """
        Reads any number of registers by splitting large requests into chunks.

        Args:
            start (int): The starting address of the registers.
            count (int): The total number of registers to read.
            max_registers_per_read (int): Maximum registers per Modbus request (default: 120).

        Returns:
            list: Combined list of all read registers.

        Raises:
            DeviceConnectionError: If reading failed
        """
        handle_connection = not self.connected
        if handle_connection:
            await self.open_connection()

        registers = []
        try:
            for offset in range(0, count, max_registers_per_read):
                chunk_start = start + offset
                remaining = count - offset
                chunk_count = min(remaining, max_registers_per_read)

                response = await self.client.read_holding_registers(
                    chunk_start, count=chunk_count, device_id=self.modbus_address
                )
                registers.extend(response.registers)

        except Exception as e:
            raise DeviceConnectionError(
                f"Failed to read holding registers [{chunk_start}-{chunk_start+count-1}] : {e}"
            ) from e
        finally:
            if handle_connection:
                await self.close_connection()

        return registers

    async def read_input_registers(self, start, count, max_registers_per_read=120):
        """
        Reads any number of input registers by splitting large requests into chunks.

        Args:
            start (int): The starting address of the registers.
            count (int): The total number of registers to read.
            max_registers_per_read (int): Maximum registers per Modbus request (default: 120).

        Returns:
            list: Combined list of all read registers.

        Raises:
            DeviceConnectionError: If reading failed
        """
        handle_connection = not self.connected
        if handle_connection:
            await self.open_connection()

        registers = []
        try:
            for offset in range(0, count, max_registers_per_read):
                chunk_start = start + offset
                remaining = count - offset
                chunk_count = min(remaining, max_registers_per_read)

                response = await self.client.read_input_registers(
                    chunk_start, count=chunk_count, device_id=self.modbus_address
                )
                registers.extend(response.registers)

        except Exception as e:
            raise DeviceConnectionError(
                f"Failed to read input registers [{chunk_start}-{chunk_start+count-1}]: {e}"
            ) from e
        finally:
            if handle_connection:
                await self.close_connection()

        return registers

    async def write_holding_register(self, address, value):
        """
        Writes a single value to a holding register.

        Args:
            address (int): Register address.
            value (int): Value to write.

        Raises:
            DeviceConnectionError: If writing failed


        """

        handle_connection = not self.connected
        if handle_connection:
            await self.open_connection()

        try:
            response = await self.client.write_register(
                address=address, value=value, device_id=self.modbus_address
            )
            return not response.isError()
        except Exception as e:
            raise DeviceConnectionError(
                f"Failed to write holding register[{address}]: {e}"
            ) from e
        finally:
            if handle_connection:
                await self.close_connection()

    async def write_holding_registers(
        self, address, values, max_registers_per_write=120
    ):
        """
        Writes multiple values to holding registers in chunks.

        Args:
            address (int): Starting register address.
            values (list[int]): List of values to write.
            max_registers_per_write (int): Maximum number of registers per write operation.

        Raises:
            DeviceConnectionError: If writing failed
        """
        handle_connection = not self.connected
        if handle_connection:
            await self.open_connection()

        try:
            total_values = len(values)
            for offset in range(0, total_values, max_registers_per_write):
                chunk_address = address + offset
                chunk_values = values[offset : offset + max_registers_per_write]

                response = await self.client.write_registers(
                    address=chunk_address,
                    values=chunk_values,
                    device_id=self.modbus_address,
                )

                if response.isError():
                    raise DeviceConnectionError(
                        f"Write error at address {chunk_address}"
                    )
        except Exception as e:
            raise DeviceConnectionError(
                f"Failed to write holding registers [{address}-{address+len(values)-1}]: {e}"
            ) from e
        finally:
            if handle_connection:
                await self.close_connection()



import logging
import time
import asyncio
import struct

from pyiskra.exceptions import InvalidResponseCode, MeasurementTypeNotSupported
from .BaseDevice import Device
from ..adapters import RestAPI, Modbus
from ..helper import (
    ModbusMapper,
    Measurements,
    Measurement,
    Phase_Measurements,
    Total_Measurements,
    Counter,
    Counters,
    get_counter_units,
    get_counter_direction,
    get_counter_type,
    IntervalMeasurementStats,
    MeasurementType,
)

log = logging.getLogger(__name__)


class WM(Device):
    """
    Represents an Impact device.

    Attributes:
        supports_measurements (bool): Indicates whether the device supports measurements.
        supports_counters (bool): Indicates whether the device supports counters.
        fw_version (float): The firmware version of the device.
    """

    DEVICE_PARAMETERS = {
        "WM3M4": {"phases": 3, "non_resettable_counters": 2},
        "WM3": {"phases": 3, "resettable_counters": 4, "non_resettable_counters": 4},
        "WM1": {"phases": 1, "resettable_counters": 4},
        # Add more models as needed
    }

    supports_measurements = True
    supports_counters = True
    supports_interval_measurements = True

    async def init(self):
        """
        Initializes the Impact device.

        This method retrieves basic information, updates the status, and logs a success message.
        """
        await self.get_basic_info()
        await self.update_status()
        log.debug(f"Successfully initialized {self.model} {self.serial}")

    async def get_measurements(
        self, measurement_type: MeasurementType = MeasurementType.ACTUAL_MEASUREMENTS
    ):
        """
        Retrieves measurements from the device.

        Returns:
            dict: A dictionary containing the measurements.
        """
        if (
            measurement_type != MeasurementType.ACTUAL_MEASUREMENTS
            and self.supports_interval_measurements == False
        ):
            raise MeasurementTypeNotSupported(
                f"{measurement_type} is not supported by {self.model}"
            )

        if isinstance(self.adapter, RestAPI):
            log.debug(
                f"Getting measurements from Rest API for {self.model} {self.serial}"
            )
            return await self.adapter.get_measurements()
        elif isinstance(self.adapter, Modbus):
            log.debug(
                f"Getting measurements from Modbus for {self.model} {self.serial}"
            )

            offset = 0
            last_interval_duration = None
            time_since_last_measurement = None
            avg_measurement_counter = None

            # Other measurement type registers are just shifted
            if measurement_type == MeasurementType.AVERAGE_MEASUREMENTS:
                offset = 5400
            elif measurement_type == MeasurementType.MAX_MEASUREMENTS:
                offset = 5500
            elif measurement_type == MeasurementType.MIN_MEASUREMENTS:
                offset = 5600

            interval_stats = None
            data = await self.adapter.read_input_registers(100 + offset, 91)
            mapper = ModbusMapper(data, 100)

            if measurement_type != MeasurementType.ACTUAL_MEASUREMENTS:
                interval_stats = IntervalMeasurementStats()
                interval_data = await self.adapter.read_input_registers(5500, 2)
                interval_stats_mapper = ModbusMapper(interval_data, 100)
                interval_stats.last_interval_duration = (
                    interval_stats_mapper.get_uint16(100) / 10
                )
                interval_stats.time_since_last_measurement = (
                    interval_stats_mapper.get_int16(101) / 10
                )

            phases = []
            for phase in range(self.phases):
                voltage = Measurement(
                    mapper.get_t5(107 + 2 * phase),
                    "V",
                )
                current = Measurement(
                    mapper.get_t5(126 + 2 * phase),
                    "A",
                )
                active_power = Measurement(
                    mapper.get_t6(142 + 2 * phase),
                    "W",
                )
                reactive_power = Measurement(
                    mapper.get_t6(150 + 2 * phase),
                    "var",
                )
                apparent_power = Measurement(
                    mapper.get_t5(158 + 2 * phase),
                    "VA",
                )
                power_factor = Measurement(
                    mapper.get_t7(166 + 2 * phase)["value"],
                    "",
                )
                power_angle = Measurement(
                    mapper.get_int16(173 + phase) / 100,
                    "°",
                )
                thd_voltage = Measurement(
                    mapper.get_uint16(182 + phase) / 100,
                    "%",
                )
                thd_current = Measurement(
                    mapper.get_uint16(188 + phase) / 100,
                    "%",
                )
                phases.append(
                    Phase_Measurements(
                        voltage,
                        current,
                        active_power,
                        reactive_power,
                        apparent_power,
                        power_factor,
                        power_angle,
                        thd_voltage,
                        thd_current,
                    )
                )

            active_power_total = Measurement(
                mapper.get_t6(140),
                "W",
            )
            reactive_power_total = Measurement(
                mapper.get_t6(148),
                "var",
            )
            apparent_power_total = Measurement(
                mapper.get_t5(156),
                "VA",
            )
            power_factor_total = Measurement(
                mapper.get_t7(164)["value"],
                "",
            )
            power_angle_total = Measurement(
                mapper.get_int16(172) / 100,
                "°",
            )
            frequency = Measurement(
                mapper.get_t5(105),
                "Hz",
            )
            temperature = Measurement(
                mapper.get_int16(181) / 100,
                "°C",
            )
            total = Total_Measurements(
                active_power_total,
                reactive_power_total,
                apparent_power_total,
                power_factor_total,
                power_angle_total,
            )

            return Measurements(phases, total, frequency, temperature, interval_stats)

    async def get_counters(self):
        """
        Retrieves counters from the device.

        Returns:
            dict: A dictionary containing the counters.
        """
        if isinstance(self.adapter, RestAPI):
            log.debug(f"Getting counters from Rest API for {self.model} {self.serial}")
            return await self.adapter.get_counters()
        elif isinstance(self.adapter, Modbus):
            log.debug(
                f"Getting measurements from Modbus for {self.model} {self.serial}"
            )

            # Open the connection
            handle_connection = not self.adapter.connected
            if handle_connection:
                await self.adapter.open_connection()

            data = await self.adapter.read_input_registers(400, 64)
            data_mapper = ModbusMapper(data, 400)
            direction_settings = await self.adapter.read_holding_registers(151, 1)
            counter_settings = await self.adapter.read_holding_registers(421, 36)
            counter_settings_mapper = ModbusMapper(counter_settings, 421)

            # Close the connection
            if handle_connection:
                await self.adapter.close_connection()

            non_resettable = []
            resettable = []
            reverse_connection = False
            if direction_settings[0] & 2:
                reverse_connection = True

            for counter in range(self.resettable_counters):
                units = get_counter_units(
                    counter_settings_mapper.get_uint16(421 + 10 * counter)
                )

                direction = get_counter_direction(
                    counter_settings_mapper.get_uint16(422 + 10 * counter),
                    reverse_connection,
                )

                counter_type = get_counter_type(direction, units)

                value = data_mapper.get_int32(406 + 2 * counter)

                exponent = data_mapper.get_int16(401 + counter)

                resettable.append(
                    Counter(
                        value * (10**exponent),
                        units,
                        direction,
                        counter_type,
                    )
                )

            for counter in range(self.non_resettable_counters):
                units = get_counter_units(
                    counter_settings_mapper.get_uint16(
                        421 + 10 * (counter + self.resettable_counters)
                    )
                )

                direction = get_counter_direction(
                    counter_settings_mapper.get_uint16(
                        422 + 10 * (counter + self.resettable_counters)
                    ),
                    reverse_connection,
                )

                counter_type = get_counter_type(direction, units)

                value = data_mapper.get_int32(
                    406 + 2 * (counter + self.resettable_counters)
                )

                exponent = data_mapper.get_int16(
                    401 + (counter + self.resettable_counters)
                )

                non_resettable.append(
                    Counter(
                        value * (10**exponent),
                        units,
                        direction,
                        counter_type,
                    )
                )

            return Counters(non_resettable, resettable)

    async def update_status(self):
        """
        Updates the status of the device.

        This method acquires a lock to ensure that only one update is running at a time.
        It retrieves measurements and counters, updates the corresponding attributes,
        and sets the update timestamp.
        """
        # If update is already running, wait for it to finish and then return
        if self.update_lock.locked():
            log.debug("Update already running for %s %s" % (self.model, self.serial))
            while self.update_lock.locked():
                await asyncio.sleep(0.1)
            return

        # If update is not running, acquire the lock and update
        async with self.update_lock:
            log.debug("Updating status for %s %s" % (self.model, self.serial))

            # if the adapter is Modbus, open the connection
            if isinstance(self.adapter, Modbus):
                await self.adapter.open_connection()

            self.measurements = await self.get_measurements()
            self.counters = await self.get_counters()

            # if the adapter is Modbus, close the connection
            if isinstance(self.adapter, Modbus):
                await self.adapter.close_connection()

            self.update_timestamp = time.time()



from datetime import datetime
import struct
import time
from enum import Enum


class IntervalMeasurementStats:
    last_interval_duration: int
    time_since_last_measurement: int


class BasicInfo:
    def __init__(
        self,
        serial,
        model,
        description,
        location,
        sw_ver,
    ):
        self.serial = serial
        self.model = model
        self.description = description
        self.location = location
        self.sw_ver = sw_ver

    def __str__(self):
        return f'{self.model} {self.serial} {self.sw_ver} {self.description} {self.location}'

class Measurement:
    value: float
    units: str

    def __init__(self, value=None, units=None):
        self.value = value
        self.units = units


class Phase_Measurements:
    voltage: Measurement
    current: Measurement
    active_power: Measurement
    reactive_power: Measurement
    apparent_power: Measurement
    power_factor: Measurement
    power_angle: Measurement
    thd_voltage: Measurement
    thd_current: Measurement

    def __init__(
        self,
        voltage=None,
        current=None,
        active_power=None,
        reactive_power=None,
        apparent_power=None,
        power_factor=None,
        power_angle=None,
        thd_voltage=None,
        thd_current=None,
    ):
        self.voltage = voltage
        self.current = current
        self.active_power = active_power
        self.reactive_power = reactive_power
        self.apparent_power = apparent_power
        self.power_factor = power_factor
        self.power_angle = power_angle
        self.thd_voltage = thd_voltage
        self.thd_current = thd_current


class Total_Measurements:
    active_power: Measurement
    reactive_power: Measurement
    apparent_power: Measurement
    power_factor: Measurement
    power_angle: Measurement

    def __init__(
        self,
        active_power=None,
        reactive_power=None,
        apparent_power=None,
        power_factor=None,
        power_angle=None,
    ):
        self.active_power = active_power
        self.reactive_power = reactive_power
        self.apparent_power = apparent_power
        self.power_factor = power_factor
        self.power_angle = power_angle


class Measurements:
    phases: list[Phase_Measurements]
    total: Total_Measurements
    frequency: Measurement
    temperature: Measurement
    interval_stats: IntervalMeasurementStats

    def __init__(
        self,
        phases=None,
        total=None,
        frequency=None,
        temperature=None,
        interval_stats=None,
    ):
        self.timestamp = time.time()

        self.phases = phases
        self.total = total
        self.frequency = frequency
        self.temperature = temperature
        self.interval_stats = interval_stats


class Consumed_Energy:
    total:Measurement
    timestamp_total: Measurement
    last_month: Measurement
    timestamp_last_month: Measurement
    two_months_ago:Measurement
    timestamp_two_months_ago: Measurement
    last_year: Measurement
    timestamp_last_year: Measurement
    two_years_ago: Measurement
    timestamp_two_years_ago: Measurement
    this_month: Measurement
    previous_month: Measurement
    this_year: Measurement
    previous_year: Measurement

    def __init__(
        self,
        total=None,
        timestamp_total=None,
        last_month=None,
        timestamp_last_month=None,
        two_months_ago=None,
        timestamp_two_months_ago=None,
        last_year=None,
        timestamp_last_year=None,
        two_years_ago=None,
        timestamp_two_years_ago=None,
        this_month=None,
        previous_month=None,
        this_year=None,
        previous_year=None,

    ):
        self.total = total
        self.timestamp_total = timestamp_total
        self.last_month = last_month
        self.timestamp_last_month = timestamp_last_month
        self.two_months_ago = two_months_ago
        self.timestamp_two_months_ago = timestamp_two_months_ago
        self.last_year = last_year
        self.timestamp_last_year = timestamp_last_year
        self.two_years_ago = two_years_ago
        self.timestamp_two_years_ago = timestamp_two_years_ago
        self.this_month = this_month
        self.previous_month = previous_month
        self.this_year = this_year
        self.previous_year = previous_year

class Excess_Power:
    excess_power_limit: Measurement
    excess_power_this_month: Measurement
    excess_power_previous_month: Measurement

    def __init__(
        self,
        excess_power_limit=None,
        excess_power_this_month=None,
        excess_power_previous_month=None,
    ):
        self.excess_power_limit = excess_power_limit
        self.excess_power_this_month = excess_power_this_month
        self.excess_power_previous_month = excess_power_previous_month


class Max_15min_Power:
    max_15min_power_since_reset: Measurement
    timestamp_since_reset: Measurement
    max_15min_power_this_month: Measurement
    timestamp_this_month: Measurement
    max_15min_power_previous_month: Measurement
    timestamp_previous_month: Measurement
    max_15min_power_this_year: Measurement
    timestamp_this_year: Measurement
    max_15min_power_previous_year: Measurement
    timestamp_previous_year: Measurement
    reset_timestamp: Measurement

    def __init__(
        self,
        max_15min_power_since_reset=None,
        timestamp_since_reset=None,
        max_15min_power_this_month=None,
        timestamp_this_month=None,
        max_15min_power_previous_month=None,
        timestamp_previous_month=None,
        max_15min_power_this_year=None,
        timestamp_this_year=None,
        max_15min_power_previous_year=None,
        timestamp_previous_year=None,
        reset_timestamp=None,
    ):
        self.max_15min_power_since_reset = max_15min_power_since_reset
        self.timestamp_since_reset = timestamp_since_reset
        self.max_15min_power_this_month = max_15min_power_this_month
        self.timestamp_this_month = timestamp_this_month
        self.max_15min_power_previous_month = max_15min_power_previous_month
        self.timestamp_previous_month = timestamp_previous_month
        self.max_15min_power_this_year = max_15min_power_this_year
        self.timestamp_this_year = timestamp_this_year
        self.max_15min_power_previous_year = max_15min_power_previous_year
        self.timestamp_previous_year = timestamp_previous_year
        self.reset_timestamp = reset_timestamp
        
class Active_Power_Measurements:
    actual_value: Measurement
    thermal_function: Measurement
    predicted_15min: Measurement
    predicted_15min_active_limit: Measurement
    last_15min: Measurement
    max_15min_since_reset: Measurement
    active_energy_total: Measurement
    timestamp: Measurement

    def __init__(
        self,
        actual_value=None,
        thermal_function=None,
        predicted_15min=None,
        predicted_15min_active_limit=None,
        last_15min=None,
        max_15min_since_reset=None,
        active_energy_total=None,
        timestamp=None,
    ):
        self.actual_value = actual_value
        self.thermal_function = thermal_function
        self.predicted_15min = predicted_15min
        self.predicted_15min_active_limit = predicted_15min_active_limit
        self.last_15min = last_15min
        self.max_15min_since_reset = max_15min_since_reset
        self.active_energy_total = active_energy_total
        self.timestamp = timestamp

class Time_Block:
    consumed_energy: Consumed_Energy
    excess_power: Excess_Power
    max_15min_Power: Max_15min_Power

    def __init__(
        self,
        consumed_energy=None,
        excess_power=None,
        max_15min_Power=None,
    ):
        self.consumed_energy = consumed_energy
        self.excess_power = excess_power
        self.max_15min_Power = max_15min_Power

class Time_Blocks_Measurements:
    time_blocks: list[Time_Block]
    active_power_measurements_import: Active_Power_Measurements
    active_power_measurements_export: Active_Power_Measurements
    active_block_index: Measurement
    time_to_end_interval: Measurement

    def __init__(
        self,
        time_blocks=None,
        active_power_measurements_import=None,
        active_power_measurements_export=None,
        active_block_index=None,
        time_to_end_interval=None,
    ):
        self.timestamp = time.time()

        self.time_blocks = time_blocks
        self.active_power_measurements_import = active_power_measurements_import
        self.active_power_measurements_export = active_power_measurements_export
        self.active_block_index = active_block_index
        self.time_to_end_interval = time_to_end_interval

    @property
    def active_time_block(self):
        return self.timeBlocks[self.active_block_index]

class CounterType(Enum):
    ACTIVE_IMPORT = "active_import"
    ACTIVE_EXPORT = "active_export"
    REACTIVE_IMPORT = "reactive_import"
    REACTIVE_EXPORT = "reactive_export"
    APPARENT_IMPORT = "apparent_import"
    APPARENT_EXPORT = "apparent_export"
    UNKNOWN = "unknown"


class MeasurementType(Enum):
    ACTUAL_MEASUREMENTS = "actual_measurements"
    AVERAGE_MEASUREMENTS = "average_measurements"
    MIN_MEASUREMENTS = "min_measurements"
    MAX_MEASUREMENTS = "max_measurements"


class Counter:
    value: float
    units: str
    direction: str
    counter_type: CounterType

    def __init__(
        self,
        value=None,
        units=None,
        direction=None,
        counter_type=None,
    ):
        self.value = value
        self.units = units
        self.direction = direction
        self.counter_type = counter_type


class Counters:
    non_resettable: list[Counter]
    resettable: list[Counter]

    def __init__(self, non_resettable=None, resettable=None):
        self.timestamp = time.time()
        self.non_resettable = non_resettable if non_resettable is not None else []
        self.resettable = resettable if resettable is not None else []


counter_units = ["", "Wh", "varh", "VAh"]

def get_counter_units(counter_parameter):
    # first 2 bits only
    return counter_units[counter_parameter & 0x3] 


def get_counter_direction(quadrants, reverse_connection):
    quadrants = quadrants & 0x0F
    direction = 0
    if quadrants == 9 or quadrants == 3:
        direction = "export"
    elif quadrants == 6 or quadrants == 12:
        direction = "import"
    elif quadrants == 15:
        direction = "bidirectional"

    if reverse_connection:
        if direction == "import":
            direction = "export"
        elif direction == "export":
            direction = "import"

    return direction


def get_counter_type(direction, units):
    if direction == "import":
        if units == "Wh":
            return CounterType.ACTIVE_IMPORT
        elif units == "varh":
            return CounterType.REACTIVE_IMPORT
        elif units == "VAh":
            return CounterType.APPARENT_IMPORT
    elif direction == "export":
        if units == "Wh":
            return CounterType.ACTIVE_EXPORT
        elif units == "varh":
            return CounterType.REACTIVE_EXPORT
        elif units == "VAh":
            return CounterType.APPARENT_EXPORT

    return CounterType.UNKNOWN


class ModbusMapper:
    def __init__(self, register_values, start_address):
        self.register_values = register_values
        self.start_address = start_address

    def get_value(self, desired_address):
        if (
            desired_address < self.start_address
            or desired_address >= self.start_address + len(self.register_values)
        ):
            raise Exception("desired address out of range")

        index = desired_address - self.start_address
        return self.register_values[index]

    def get_t5(self, desired_address, word_swap=False):
        high_word = self.get_value(desired_address)
        low_word = self.get_value(desired_address + 1)

        combined = (high_word << 16) | low_word
        if word_swap:
            combined = (low_word << 16) | high_word
        value = combined & 0xFFFFFF  # bits 0-23
        exponent = (combined >> 24) & 0xFF  # bits 24-31
        if exponent & 0x80:  # if the sign bit is set
            exponent -= 0x100  # convert to signed
        return round(value * (10**exponent), 3)

    def get_t6(self, desired_address, word_swap=False):
        high_word = self.get_value(desired_address)
        low_word = self.get_value(desired_address + 1)

        combined = (high_word << 16) | low_word
        if word_swap:
            combined = (low_word << 16) | high_word
        value = combined & 0xFFFFFF  # bits 0-23
        exponent = (combined >> 24) & 0xFF  # bits 24-31
        if exponent & 0x80:  # if the sign bit is set
            exponent -= 0x100  # convert to signed
        if value & 0x800000:  # if the sign bit is set
            value -= 0x1000000  # convert to signed
        return round(value * (10**exponent), 3)

    def get_t7(self, desired_address, word_swap=False):
        high_word = self.get_value(desired_address)
        low_word = self.get_value(desired_address + 1)

        combined = (high_word << 16) | low_word
        if word_swap:
            combined = (low_word << 16) | high_word

        value = combined & 0xFFFF  # bits 0-15
        inductive_capacitive = (combined >> 16) & 0xFF  # bits 16-23
        import_export = (combined >> 24) & 0xFF  # bits 24-31

        inductive_capacitive_str = (
            "inductive" if inductive_capacitive == 0x00 else "capacitive"
        )
        import_export_str = "import" if import_export == 0x00 else "export"

        return {
            "value": value / 10000,
            "inductive_capacitive": inductive_capacitive_str,
            "import_export": import_export_str,
        }

    def get_uint16(self, desired_address):
        value = self.get_value(desired_address)
        return value

    def get_int16(self, desired_address):
        value = self.get_value(desired_address)
        if value is None:
            return None
        if value > 32767:
            value -= 65536
        return value

    def get_timestamp(self, desired_address):
        value = self.get_uint32(desired_address)
        return time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime(value))

    def get_float(self, desired_address, word_swap=False):
        value = self.get_value(desired_address)
        next_value = self.get_value(desired_address + 1)
        if value is None or next_value is None:
            return None
        # Combine the two 16-bit register values into a 32-bit integer
        combined = (value << 16) | next_value

        if word_swap:
            combined = (next_value << 16) | value
        # Convert the 32-bit integer to a float
        return round(struct.unpack("!f", struct.pack("!I", combined))[0], 3)

    def get_uint32(self, desired_address, word_swap=False):
        high_word = self.get_value(desired_address)
        low_word = self.get_value(desired_address + 1)
        if word_swap:
            return (low_word << 16) + high_word
        return (high_word << 16) + low_word

    def get_int32(self, desired_address, word_swap=False):
        high_word = self.get_value(desired_address)
        low_word = self.get_value(desired_address + 1)
        value = (high_word << 16) + low_word
        if word_swap:
            value = (low_word << 16) + high_word

        if value & 0x80000000:  # if the sign bit is set
            value -= 0x100000000  # convert to signed
        return value

    def get_string(self, desired_address):
        value = self.get_value(desired_address)
        high_byte = (value >> 8) & 0xFF
        low_byte = value & 0xFF
        return "".join([chr(high_byte), chr(low_byte)])

    def get_string_range(self, desired_address, size):
        return "".join(
            [
                self.get_string(register)
                for register in range(desired_address, desired_address + size)
            ]
        )

    def dump(self):
        for i, value in enumerate(self.register_values):
            print(f"Address {self.start_address + i}: {value} 0x{value:04X}")



# Description: Custom exceptions for pyiskra


class NotAuthorised(Exception):
    pass


class DeviceNotSupported(Exception):
    pass


class ProtocolNotSupported(Exception):
    pass


class DeviceConnectionError(Exception):
    pass


class DeviceTimeoutError(Exception):
    pass


class InvalidResponseCode(Exception):
    pass


class MeasurementTypeNotSupported(Exception):
    pass



import asyncio
import socket
import re
from ..exceptions import DeviceConnectionError, DeviceNotSupported, ProtocolNotSupported
from ..adapters import RestAPI


class Device:
    """
    Represents a base device in the Iskra system.

    Attributes:
        model (str): The model of the device.
        serial (str): The serial number of the device.
        description (str): The description of the device.
        location (str): The location of the device.
        supports_measurements (bool): Indicates whether the device supports measurements.
        supports_counters (bool): Indicates whether the device supports counters.
        measurements (None or dict): The measurements data of the device.
        counters (None or dict): The counters data of the device.
        update_timestamp (int): The timestamp of the last update.
    """

    DEVICE_PARAMETERS = {}
    model = ""
    serial = ""
    description = ""
    location = ""
    is_gateway = False
    supports_measurements = False
    supports_counters = False
    supports_time_blocks = False
    measurements = None
    counters = None
    phases = 0
    current_nominal = 0
    voltage_nominal = 0
    non_resettable_counters = 0
    resettable_counters = 0
    fw_version = None
    parent_device = None

    update_timestamp = 0

    @staticmethod
    async def create_device(adapter, parent_device=None):
        """
        Creates a device based on the adapter.

        Args:
            adapter: The adapter used to communicate with the device.

        Returns:
            An instance of the appropriate device subclass based on the model.

        Raises:
            ConnectionError: If failed to get basic info from the adapter.
            ValueError: If the device model is unsupported.
        """

        basic_info = await adapter.get_basic_info()

        model = basic_info.model

        from .IMPACT import Impact

        if any(
            re.match(model_pattern, model)
            for model_pattern in Impact.DEVICE_PARAMETERS.keys()
        ):
            return Impact(adapter, parent_device)

        from .WM import WM

        if any(
            re.match(model_pattern, model)
            for model_pattern in WM.DEVICE_PARAMETERS.keys()
        ):
            return WM(adapter, parent_device)

        from .MeasuringCenter import MeasuringCentre

        if any(
            re.match(model_pattern, model)
            for model_pattern in MeasuringCentre.DEVICE_PARAMETERS.keys()
        ):
            return MeasuringCentre(adapter, parent_device)

        from .SmartGateway import SmartGateway

        if any(
            re.match(model_pattern, model)
            for model_pattern in SmartGateway.DEVICE_PARAMETERS.keys()
        ):
            if isinstance(adapter, RestAPI):
                return SmartGateway(adapter, parent_device)
            else:
                # only REST API is supported for SmartGateway
                raise ProtocolNotSupported(f"Unsupported device model: {model}")

        raise DeviceNotSupported(f"Unsupported device model: {model}")

    def __init__(self, adapter, parent_device=None):
        """
        Initializes the Iskra Device.

        Args:
            adapter: The adapter used to communicate with the device.
        """
        self.adapter = adapter
        self.update_lock = asyncio.Lock()
        self.parent_device = parent_device

    async def get_basic_info(self):
        """
        Retrieves basic information from the device.

        Returns:
            dict: A dictionary containing the basic information.
        """
        basic_info = await self.adapter.get_basic_info()

        self.serial = basic_info.serial
        self.model = basic_info.model
        self.description = basic_info.description

        # Use regular expressions to match the model name and assign parameters accordingly
        for model_pattern, parameters in self.DEVICE_PARAMETERS.items():
            if re.match(model_pattern, self.model):
                self.phases = parameters["phases"]
                self.resettable_counters = parameters.get("resettable_counters", 0)
                self.non_resettable_counters = parameters.get("non_resettable_counters", 0)
                break

        return basic_info

    async def update_status(self):
        """
        Updates the status of the device.

        This method needs to be re-defined in all sub-classes.
        """
        raise NotImplementedError

    async def init(self):
        """
        Initializes the status of the device.

        This method needs to be re-defined in all sub-classes.
        """
        raise NotImplementedError

    @property
    def ip_address(self):
        """
        Returns the IP address of the device.

        Returns:
            The IP address of the device.
        """
        return self.adapter.ip_address



from .BaseDevice import Device
from ..adapters.RestAPI import RestAPI
from ..exceptions import DeviceNotSupported, DeviceConnectionError


import logging

log = logging.getLogger(__name__)


class SmartGateway(Device):
    """
    Represents a smart gateway device.

    Attributes:
        supports_measurements (bool): Indicates if the device supports measurements.
        supports_counters (bool): Indicates if the device supports counters.
        serial (str): The serial number of the device.
        model (str): The model of the device.
        description (str): The description of the device.
        location (str): The location of the device.
        fw_version (str): The firmware version of the device.
        child_devices (list): A list of child devices connected to the gateway.
    """

    DEVICE_PARAMETERS = {
        "SG": {"phases": 0}
    }

    is_gateway = True
    child_devices = []
    supports_measurements = False
    supports_counters = False

    async def init(self):
        """
        Initializes the smart gateway device.

        This method retrieves basic information about the device, such as serial number, model, description, location,
        and firmware version. It also initializes the child devices connected to the gateway.
        """
        basic_info = await self.adapter.get_basic_info()
        self.serial = basic_info.serial
        self.model = basic_info.model
        self.description = basic_info.description
        self.location = basic_info.location
        self.fw_version = basic_info.sw_ver
        await self.update_child_devices()
        log.debug(f"Successfully initialized {self.model} {self.serial}")

    async def update_child_devices(self):
        self.child_devices = []
        child_devices = await self.adapter.get_devices()

        child_devices = [
            device for device in child_devices if "Right IR" not in device["interface"]
        ]
        for i, device in enumerate(child_devices):
            if device["model"] != "Disabled" and device["model"] != "Not Detected":
                log.debug(
                    f"Found device {device['model']} {device['serial']} connected to {self.model} {self.serial}"
                )
                adapter = RestAPI(
                    ip_address=self.adapter.ip_address,
                    authentication=self.adapter.authentication,
                    device_index=i,
                )
                try:
                    dev = await Device.create_device(adapter, self)
                    await dev.init()
                    self.child_devices.append(dev)
                except DeviceNotSupported as e:
                    log.error(
                        f"Failed to create device {device['model']} {device['serial']}: {e}"
                    )
                except DeviceConnectionError as e:
                    log.error(
                        f"Failed to connect to device {device['model']} {device['serial']}: {e}"
                    )

    def get_child_devices(self):
        """
        Returns the list of child devices connected to the gateway.

        Returns:
            list: A list of child devices.
        """
        return self.child_devices

    async def get_measurements(self):
        """
        Retrieves the measurements from the smart gateway.

        Returns:
            dict: A dictionary containing the measurements.
        """
        return await self.adapter.get_measurements()

    async def get_counters(self):
        """
        Retrieves the counters from the smart gateway.

        Returns:
            dict: A dictionary containing the counters.
        """
        return await self.adapter.get_counters()

    async def update_status(self):
        """
        Updates the status of the smart gateway.
        """
        log.debug(f"Updating status for {self.model} {self.serial}")
        # await self.update_child_devices()
